from dataclasses import dataclass, field
from typing import Any, Optional, Tuple, Union

from tree import RootedTree, RootedTreeNode


@dataclass
class PseudoTreeRoot:
    rootPaths: list[Any] = field(default_factory=list)


@dataclass
class BinaryNode:
    pathUp: "CompactedPath"
    pathLeft: "CompactedPath"
    pathRight: "CompactedPath"
    isFake: bool
    isMarked: bool
    gap: int = 0


@dataclass
class ImplicitNode:
    idxOnPath: int
    path: "CompactedPath" = field(default=None)
    markedUp: "ImplicitNode" = field(default=None)
    markedDown: "ImplicitNode" = field(default=None)


@dataclass
class CompactedPath:
    length: int = 0
    topGap: int = 0
    highestMarked: Optional[ImplicitNode] = field(default=None)
    lowestMarked: Optional[ImplicitNode] = field(default=None)
    nodeUp: Union[BinaryNode, PseudoTreeRoot] = field(default=None)
    nodeDown: Union[BinaryNode, "Leaf"] = field(default=None)
    top: Optional[ImplicitNode] = field(default=None)  # None iff path is empty
    bottom: Optional[ImplicitNode] = field(default=None)  # None iff path is empty


@dataclass
class Leaf:
    gap: int = 0
    pathUp: CompactedPath = field(default=None)


PseudotreeNode = Union[Leaf, ImplicitNode, BinaryNode, PseudoTreeRoot]


def dist(u: ImplicitNode, v: ImplicitNode) -> int:
    assert u.path == v.path, "dist only supports nodes coming from the same CompactedPath"
    return abs(u.idxOnPath - v.idxOnPath)


def unmark(v: Union[PseudotreeNode, BinaryNode, ImplicitNode]) -> int:
    localMaxGap = 0

    if isinstance(v, BinaryNode):
        v.isMarked = False
        localMaxGap = walkAndUpdate(v)

    elif isinstance(v, ImplicitNode):
        markedUp = v.markedUp
        markedDown = v.markedDown
        v.markedUp = v.markedDown = None

        if markedUp is not None:
            markedUp.markedDown = markedDown

        if markedDown is not None:
            markedDown.markedUp = markedUp

        path = v.path

        if path.lowestMarked == v:
            path.lowestMarked = markedUp
            if markedUp is not None:
                gap = path.nodeDown.gap + dist(markedUp, path.bottom)
                localMaxGap = max(localMaxGap, gap)

        if path.highestMarked == v:
            path.highestMarked = markedDown

            if path.highestMarked is not None:
                gap = dist(path.highestMarked, path.top)
            else:
                gap = path.length + path.nodeDown.gap

            path.topGap = gap
            localMaxGap = max(localMaxGap, gap, walkAndUpdate(path.nodeUp))

        if markedUp is not None and markedDown is not None:
            localMaxGap = max(localMaxGap, dist(markedUp, markedDown) - 1)

    else:
        assert isinstance(v, PseudoTreeRoot)

    return localMaxGap


def calcGapBinaryNode(v: BinaryNode) -> int:
    if v.isMarked:
        return 0
    else:
        return min(v.pathLeft.topGap, v.pathRight.topGap) + (0 if v.isFake else 1)


def walkAndUpdate(v: Union[BinaryNode, PseudoTreeRoot]) -> int:
    localMaxGap = 0

    while not isinstance(v, PseudoTreeRoot) and v.gap < calcGapBinaryNode(v):
        v.gap = calcGapBinaryNode(v)
        path = v.pathUp

        if path.lowestMarked is not None:
            gap = v.gap + 1 + dist(path.bottom, path.lowestMarked)
            localMaxGap = max(localMaxGap, gap)
            break
        else:
            path.topGap = v.gap + path.length
            localMaxGap = max(localMaxGap, path.topGap)
            v = path.nodeUp

    return localMaxGap


def createBinaryNode(pathLeft: CompactedPath, pathRight: CompactedPath, isFake: bool) -> BinaryNode:
    pathUp = CompactedPath()

    node = BinaryNode(
        pathUp=pathUp, pathLeft=pathLeft, pathRight=pathRight, isFake=isFake, isMarked=not isFake
    )

    pathUp.nodeDown = pathLeft.nodeUp = pathRight.nodeUp = node

    return node


def convertDFS(v: RootedTreeNode, ids_to_nodes: list[Optional[PseudotreeNode]]) -> CompactedPath:
    v_id = v.id

    if len(v.children) == 0:
        v = Leaf()
        ids_to_nodes[v_id] = v

        path = CompactedPath(topGap=0, nodeDown=v)
        v.pathUp = path

        return path

    elif len(v.children) == 1:
        path = convertDFS(v.children[0][0], ids_to_nodes)
        idx_on_path = 0 if path.top is None else path.top.idxOnPath + 1

        v = ImplicitNode(path=path, markedDown=path.top, idxOnPath=idx_on_path)
        ids_to_nodes[v_id] = v

        if path.bottom is None:
            path.bottom = path.lowestMarked = v

        if path.top is not None:
            path.top.markedUp = v

        path.top = path.highestMarked = v
        path.length += 1

        return path

    else:
        pathsDown = []

        for v, _ in v.children:
            path = convertDFS(v, ids_to_nodes)
            pathsDown.append(path)

        while len(pathsDown) > 2:
            leftPath = pathsDown.pop()
            rightPath = pathsDown.pop()
            fakeNode = createBinaryNode(leftPath, rightPath, isFake=True)
            pathsDown.append(fakeNode.pathUp)

        v = createBinaryNode(pathLeft=pathsDown[0], pathRight=pathsDown[1], isFake=False)
        ids_to_nodes[v_id] = v

        return v.pathUp


def convert(T: RootedTree) -> Tuple[PseudoTreeRoot, list[PseudotreeNode]]:
    pseudoTreeRoot = PseudoTreeRoot()
    ids_to_nodes = [None for _ in range(T.size)]

    for (v, _) in T.root.children:
        path = convertDFS(v, ids_to_nodes)
        path.nodeUp = pseudoTreeRoot
        pseudoTreeRoot.rootPaths.append(path)

    ids_to_nodes[T.root.id] = pseudoTreeRoot

    for idx, node in enumerate(ids_to_nodes):
        assert node is not None, f"Node {idx} is not accessible from root"

    return pseudoTreeRoot, ids_to_nodes


def solve_directed_cover(T: RootedTree) -> list[Tuple[int, int]]:
    # Picking a candidate L and precalculating TreePref
    (L, ids) = T.get_path_from_leaf()
    treepref = T.compute_treepref(L)

    # Converting T into PseudoTree
    pseudotree_root, ids_to_pseudotree_nodes = convert(T)

    # Grouping PseudoTree nodes by its treepref
    nodes_by_treepref = [[] for _ in range(len(L) + 1)]
    min_treepref_for_leaf = len(L) + 1
    for tree_pref, node in zip(treepref, ids_to_pseudotree_nodes):
        nodes_by_treepref[tree_pref].append(node)
        if isinstance(node, Leaf):
            min_treepref_for_leaf = min(min_treepref_for_leaf, tree_pref)

    # Calculating covers
    maxGap = 0
    covers = []

    for k in range(min_treepref_for_leaf):
        # Unmarking nodes having treepref_L == k
        for v in nodes_by_treepref[k]:
            maxGap = max(maxGap, unmark(v))

        if maxGap <= k:
            covers.append((ids[0], ids[k + 1]))

    return covers
