from dataclasses import dataclass, field
from typing import Any, Optional, Tuple

Character = Any
String = list[Character]


@dataclass
class RootedTreeNode:
    id: int
    height: int = field(default=0)
    parent: Optional["RootedTreeNode"] = field(default=None)
    labelToParent: Optional[Character] = field(default=None)
    children: list[tuple["RootedTreeNode", Character]] = field(default_factory=list)


@dataclass
class RootedTree:
    size: int
    root: RootedTreeNode
    nodes: list[RootedTreeNode] = field(default_factory=list)

    # naive implementation in O(n^2)
    def compute_treepref(self, W: String) -> list[int]:
        treepref = [0 for _ in range(self.size)]

        def lcp(w1: String, w2: String) -> int:
            res = 0

            for c1, c2 in zip(w1, w2):
                if c1 != c2:
                    break
                res += 1

            return res

        def dfs(node: RootedTreeNode, string_to_root: String) -> None:
            treepref[node.id] = lcp(string_to_root[::-1], W)

            for (child, label) in node.children:
                string_to_root.append(label)
                dfs(child, string_to_root)
                string_to_root.pop()

        dfs(self.root, [])

        return treepref

    def get_path_from_leaf(self) -> Tuple[String, list[int]]:
        for node in self.nodes:
            if len(node.children) == 0:
                return self.get_path_to_root(node.id)

        assert False, "No leaves found."

    def get_path_to_root(self, id) -> Tuple[String, list[int]]:
        node = self.nodes[id]
        string = []
        ids = []

        while node.id != self.root.id:
            string.append(node.labelToParent)
            ids.append(node.id)
            node = node.parent
        ids.append(self.root.id)

        return string, ids


@dataclass
class Node:
    id: int
    neighbours: list[tuple["Node", Character]] = field(default_factory=list)


@dataclass
class Tree:
    nodes: list[Node] = field(default_factory=list)

    def get_any_leaf(self) -> int:
        for node in self.nodes:
            if len(node.neighbours) == 1:
                return node.id

        assert False, "No leaves found."

    def root_tree(self, root_id: int) -> RootedTree:
        visited = [False for _ in range(len(self.nodes))]

        rooted_nodes = [None for _ in range(len(self.nodes))]

        def dfs(node: Node) -> RootedTreeNode:
            visited[node.id] = True
            ret = RootedTreeNode(id=node.id)
            rooted_nodes[node.id] = ret

            for (w, label) in node.neighbours:
                if not visited[w.id]:
                    w = dfs(w)
                    ret.children.append((w, label))
                    w.parent = ret
                    w.labelToParent = label
                    ret.height = max(ret.height, w.height + 1)

            return ret

        root = dfs(self.nodes[root_id])

        for idx, node in enumerate(rooted_nodes):
            assert node is not None, f"Node {idx} is not accessible from root."

        return RootedTree(size=len(self.nodes), nodes=rooted_nodes, root=root)
