from dataclasses import dataclass, field
from typing import Optional, Tuple

from tree import RootedTree, RootedTreeNode, String, Tree


@dataclass
class LinkedList:
    @dataclass
    class LinkedListNode:
        content: int
        next: Optional["LinkedListNode"] = field(default=None)

    _head: Optional[LinkedListNode] = field(default=None)
    _tail: Optional[LinkedListNode] = field(default=None)

    @staticmethod
    def create(l: list[int]) -> "LinkedList":
        ret = LinkedList()
        for v in l:
            ret.add(v)

        return ret

    def clear(self) -> None:
        self._head = self._tail = None

    def clear_but_one(self) -> None:
        if self._head is not None:
            self._head.next = None
            self._tail = self._head

    def get_front(self) -> int:
        assert self._head is not None, "List is empty"
        return self._head.content

    def empty(self) -> bool:
        return self._head is None

    def to_list(self) -> list[int]:
        node = self._head
        ret = []

        while node is not None:
            ret.append(node.content)
            node = node.next

        return ret

    def add(self, elem: int) -> None:
        if self._tail is None:
            self._head = self._tail = LinkedList.LinkedListNode(content=elem)
        else:
            node = LinkedList.LinkedListNode(content=elem)
            self._tail.next = node
            self._tail = node

    def link(self, rhs: "LinkedList") -> None:
        if rhs._head is not None:
            if self._head is None:
                assert self._tail is None, "Internal error"
                self._head, self._tail = rhs._head, rhs._tail
            else:
                self._tail.next = rhs._head
                self._tail = rhs._tail

            rhs._head = rhs._tail = None


@dataclass
class FrontVector:
    offset: int = field(default=1)
    chunk: list[Optional[LinkedList]] = field(default_factory=lambda: [None])
    # capacity is redundant in python, we can use len(chunk) instead

    def __getitem__(self, idx: int) -> LinkedList:
        assert idx >= 0
        return self.chunk[idx + self.offset]

    def __setitem__(self, idx: int, value: LinkedList) -> None:
        assert idx >= 0
        self.chunk[idx + self.offset] = value

    def push_front(self, item: LinkedList) -> None:
        current_size = len(self.chunk)

        if self.offset == 0:
            new_chunk = [None for _ in range(current_size * 2)]

            for i in range(current_size):
                new_chunk[current_size + i] = self.chunk[i]

            self.offset = current_size
            self.chunk = new_chunk

        self.offset -= 1
        self.chunk[self.offset] = item

    def size(self) -> int:
        return len(self.chunk) - self.offset


@dataclass
class GlobalContext:
    W: String
    treepref_W: list[int]
    treepref_WR: list[int]
    marked_paths: list[tuple[int, int]] = field(default_factory=list)

    def clearA(self, A: FrontVector, i: int) -> None:
        if not A[i].empty():
            val = A[i].get_front()
            if self.treepref_W[val] < i:
                A[i].clear()

    def clearB(self, B: FrontVector, i: int) -> None:
        if not B[i].empty():
            val = B[i].get_front()
            if self.treepref_WR[val] < i:
                B[i].clear()

    def matchAndMark(
        self, v: RootedTreeNode, A: FrontVector, B: FrontVector, i: int, j: int
    ) -> None:
        self.clearA(A, i)
        self.clearB(B, j)

        if not A[i].empty() and not B[j].empty():
            for u_id in A[i].to_list() + B[j].to_list():
                self.marked_paths.append((u_id, v.id))

            A[i].clear_but_one()
            B[j].clear_but_one()

    def findMatches(
        self,
        v: RootedTreeNode,
        A: FrontVector,
        B: FrontVector,
        Ap: FrontVector,
        Bp: FrontVector,
        height: int,
    ):
        lower = max(len(self.W) - A.size() + 1, 0)
        upper = min(height, len(self.W)) + 1

        for i in range(lower, upper):
            self.matchAndMark(v, A, Bp, len(self.W) - i, i)
            self.matchAndMark(v, Ap, B, i, len(self.W) - i)

    def calcAB(self, v: RootedTreeNode) -> Tuple[FrontVector, FrontVector]:
        if len(v.children) == 0:
            A = FrontVector()
            B = FrontVector()
            A.push_front(LinkedList.create([v.id]))
            B.push_front(LinkedList.create([v.id]))

        else:
            u, _ = v.children[0]

            for w, _ in v.children:
                if w.height > u.height:
                    u = w

            A, B = self.calcAB(u)
            A.push_front(LinkedList.create([v.id]))
            B.push_front(LinkedList.create([v.id]))

            for (y, _) in v.children:
                if y.id == u.id:
                    continue

                Ap, Bp = self.calcAB(y)
                Ap.push_front(LinkedList())
                Bp.push_front(LinkedList())

                self.findMatches(v, A, B, Ap, Bp, y.height + 1)

                for i in range(y.height + 2):
                    self.clearA(A, i)
                    self.clearA(Ap, i)
                    A[i].link(Ap[i])

                    self.clearB(B, i)
                    self.clearB(Bp, i)
                    B[i].link(Bp[i])

        if v.height >= len(self.W):
            self.matchAndMark(v, A, B, 0, len(self.W))
            self.matchAndMark(v, A, B, len(self.W), 0)

        return A, B


def check_if_all_marked(T: RootedTree, marked_paths: list[tuple[int, int]]) -> bool:
    cnt = [0 for _ in range(T.size)]

    for (u, v) in marked_paths:
        cnt[v] -= 1
        cnt[u] += 1

    def dfs(node: RootedTreeNode) -> int:
        for w, _ in node.children:
            cnt[node.id] += dfs(w)
        return cnt[node.id]

    dfs(T.root)

    for (idx, val) in enumerate(cnt):
        if idx != T.root.id and val == 0:
            return False

    return True


def get_candidates(T: RootedTree) -> list[Tuple[int, String]]:
    candidates = []

    def dfs(node: RootedTreeNode, string: String) -> None:
        if string:
            candidates.append((node.id, string.copy()))
        for (v, label) in node.children:
            string.append(label)
            dfs(v, string)
            string.pop()

    dfs(T.root, [])

    return candidates


def solve_undirected_cover(tree: Tree) -> list[Tuple[int, int]]:
    rootedTree = tree.root_tree(root_id=tree.get_any_leaf())

    covers = []

    for (endpoint, candidate) in get_candidates(rootedTree):
        ctx = GlobalContext(
            W=candidate,
            treepref_W=rootedTree.compute_treepref(candidate),
            treepref_WR=rootedTree.compute_treepref(candidate[::-1]),
        )

        ctx.calcAB(rootedTree.root)

        if check_if_all_marked(rootedTree, ctx.marked_paths):
            covers.append((endpoint, rootedTree.root.id))
            covers.append((rootedTree.root.id, endpoint))

    return covers
